//=== File Prolog ============================================================
//	This code was developed by RiverSoftAVG.
//
//--- Notes ------------------------------------------------------------------
//
//--- Development History  ---------------------------------------------------
//
//	03/2001	   T. Grubb
//		   Initial version.
//                 Known Bugs:
//                 - TFormationFlock.Execute command does not interpolate properly
//                   when rotating to new desiredvelocity.  It oscillates around
//                   desired velocity
//
//      File Contents:
//           TFormationBoid        One Boid (Member) of a flock that can be in
//                                 formation
//           TFormationFlock       Flock of boids in formation
//           TFormationFlockCollection Collection class to hold formation boids
//
//--- Warning ----------------------------------------------------------------
//	This software is property of RiverSoftAVG. Unauthorized use or
//      duplication of this software is strictly prohibited. Authorized users
//      are subject to the following restrictions:
//	*	RiverSoftAVG is not responsible for
//		any consequence of the use of this software.
//	*	The origin of this software must not be misrepresented either by
//		explicit claim or by omission.
//	*	Altered versions of this software must be plainly marked as such.
//	*	This notice may not be removed or altered.
//
//      This software is freeware.  You are authorized to duplicate and modify
//      this software subject to the restrictions above.
//
//=== End File Prolog ========================================================
unit FormationFlock;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Flock,
  Vector;

const
  SeparationMagnitude = 3;

type
  TFormationBoid = class(TBoid)
  { Purpose: To create a boid than can move in formation within a flock }
  private
    { Private declarations }
    FOffset: TVector;
    FOrientationOffset: TVector;
    FDistanceSqrFromOffset: Double;
    procedure SetOffset(const Value: TVector);
    procedure SetOrientationOffset(const Value: TVector);
  protected
    { Protected declarations }
    procedure ApplyCohesion( const Accumulator: TVector ); override;
    procedure ApplySeparation( const Accumulator: TVector ); override;
    procedure UpdateVelocity( const Accumulator: TVector ); override;
  public
    { Public declarations }
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure Execute; override;
    property DistanceSqrFromOffset: Double read FDistanceSqrFromOffset;
  published
    { Published declarations }
    property Offset: TVector read FOffset write SetOffset;
    property OrientationOffset: TVector read FOrientationOffset write SetOrientationOffset;
  end; { TFormationBoid }

  TFormationFlock = class;
  TFormationFlockCollection = class(TFlockCollection)
  { Purpose: To encapsulate a collection of boids that can move in formation }
  private
    { Private declarations }
    function GetItem(Index: Integer): TFormationBoid;
    procedure SetItem(Index: Integer; const Value: TFormationBoid);
    function GetFlock: TFormationFlock;
  protected
    { Protected declarations }
    property Flock: TFormationFlock read GetFlock;
  public
    { Public declarations }
    property Items[Index: Integer]: TFormationBoid read GetItem write SetItem; default;
  end; { TFormationFlockCollection }

  TFlockFormation = (ffUserDefined, ffLine, ffSquare, ffTwoSquares, ffWedge);
  TFormationFlock = class(TFlock)
  { Purpose: To encapsulate a flock which can move in formation }
  private
    { Private declarations }
    FFormation: TFlockFormation;
    FOnDefineFormation: TNotifyEvent;
    FUpdateCount: Integer;
    FDesiredPosition: TVector;
    FDesiredVelocity: TVector;
    procedure SetFormation(const Value: TFlockFormation);
    function GetBoids: TFormationFlockCollection;
    procedure SetBoids(const Value: TFormationFlockCollection);
  protected
    { Protected declarations }
    procedure FormationChange; dynamic;
    function CreateFlock: TFlockCollection; override;
    procedure DefineLineFormation; virtual;
    procedure DefineSquareFormation; virtual;
    procedure DefineSquare( Index: Integer; Count: Integer; const OffsetPos, Vel: TVector; Cols: Integer );
    procedure DefineTwoSquaresFormation; virtual;
    procedure DefineWedgeFormation; virtual;
    procedure SetDesiredPosition(const Value: TVector); virtual;
    procedure SetDesiredVelocity(const Value: TVector); virtual;
    procedure SetSeparationDistance(const Value: Single); override;
    procedure VectorChange( Sender: TVector ); override;
  public
    { Public declarations }
    procedure Assign( Source: TPersistent ); override;
    procedure AfterConstruction; override;
    constructor Create( AOwner: TComponent ); override;
    destructor Destroy; override;
    procedure DefineFormation; virtual;
    procedure Execute; override;
    procedure BeginFormationUpdate; virtual;
    procedure EndFormationUpdate; virtual;
  published
    { Published declarations }
    property Boids: TFormationFlockCollection read GetBoids write SetBoids;
    property DesiredPosition: TVector read FDesiredPosition write SetDesiredPosition;
    property DesiredVelocity: TVector read FDesiredVelocity write SetDesiredVelocity;
    property Formation: TFlockFormation read FFormation write SetFormation default ffLine;
    property OnDefineFormation: TNotifyEvent read FOnDefineFormation write FOnDefineFormation;
  end; { TFormationFlock }


procedure Register;

implementation

uses
    Math;

procedure Register;
begin
  RegisterComponents('AI', [TFormationFlock]);
end;

{ TFormationBoid }

procedure TFormationBoid.ApplyCohesion(const Accumulator: TVector);
var
   Change: TVector;
begin
     // Flocking Rule: Cohesion
     // Try to go toward our offset position from the flock's center position
     // Copy the position of center of flock
     Change := TVector.Create( Flock.Position );
     try
        // Offset
        Change.Add( Offset );
        // Calculate the distance to our desired position and save it for other routines
        FDistanceSqrFromOffset := Change.DistanceSqrTo( Position );
        Change.Subtract(Position);
        // normalize change to look more natural
        Change.SetMagnitudeOfVector(Flock.MaxUrgency*Flock.StrengthCohesion);
        // Add Change to Accumulator
        Accumulator.Add( Change );
     finally
        Change.Free;
     end;
end;

procedure TFormationBoid.Assign(Source: TPersistent);
begin
     if Source is TFormationBoid then
     begin
          FOffset.Assign( TFormationBoid( Source ).Offset );
          FOrientationOffset.Assign( TFormationBoid( Source ).OrientationOffset );
          inherited Assign(Source);
     end
     else
         inherited Assign(Source);
end;

constructor TFormationBoid.Create(Collection: TCollection);
begin
     inherited Create(Collection);
     FOffset := TVector.Create;
     FOrientationOffset := TVector.Create;
     if Flock is TFormationFlock then
        TFormationFlock(Flock).FormationChange;
end;

destructor TFormationBoid.Destroy;
var
   FormationFlock: TFormationFlock;
begin
     if Flock is TFormationFlock then
        FormationFlock := TFormationFlock(Flock)
     else
         FormationFlock := nil;
     FOffset.Free;
     FOrientationOffset.Free;
     inherited Destroy;
     if FormationFlock <> nil then
        FormationFlock.FormationChange;
end;

procedure TFormationBoid.SetOffset(const Value: TVector);
begin
     FOffset.Assign( Value );
     Changed(False);
end;

procedure TFormationBoid.SetOrientationOffset(const Value: TVector);
begin
     FOrientationOffset.Assign( Value );
     Changed(False);
end;

procedure TFormationBoid.Execute;
{ The update method is where all the actual flocking is done.  It is called
  every time period and the boid updates its position based on its flocking
  behaviors }
begin
     inherited Execute;
     // if we are close to our position then stop
     if DistanceSqrFromOffset < (Flock.MaxChange*Flock.MaxChange) then
     begin
          Speed := 0;
          Velocity.SetMagnitudeOfVector(Speed);
     end;
end;

procedure TFormationBoid.UpdateVelocity(const Accumulator: TVector);
var
   Dist: Single;
begin
     inherited UpdateVelocity(Accumulator);
     // if we are almost to our destination, start slowing down
     Dist := Sqrt(FDistanceSqrFromOffset);
     if (Dist/Flock.MaxChange) < Flock.MinTimeToMaxSpeed then
        Velocity.Multiply(Flock.MaxSpeed-(Flock.MaxChange/Dist));
end;

procedure TFormationBoid.ApplySeparation(const Accumulator: TVector);
var
   Ratio: Single;
   Change: TVector;
begin
     // we recode this rule to make sure we don't move away when we are in formation
     if Flock.StrengthSeparation = 0 then Exit;
     // Flocking Rule: Separation
     // Try to avoid boid colliding with its flockmates
     Change := TVector.Create( NearestFlockMate.Position );
     try
        // compute vector towards our nearest flock mate
        Change.Subtract( Position );
        // Compute ratio of nearest flock mate to separation distance
        // and ensure it is in range of MinUrgency..MaxUrgency
        Ratio := NearestFlockMateDist / Flock.SeparationDistance;
        if Ratio < Flock.MinUrgency then Ratio := Flock.MinUrgency
        else if Ratio > Flock.MaxUrgency then Ratio := Flock.MaxUrgency;

        Ratio := Ratio * Flock.StrengthSeparation;
        // Are we too close to nearest flockmate?  Then Move Away
        if NearestFlockMateDist < Flock.SeparationDistance*1.5 then
           Change.SetMagnitudeOfVector(-Ratio)
        // Are we too far from nearest flockmate?  Then Move Closer
        else if NearestFlockMateDist > Flock.SeparationDistance*1.5 then
           Change.SetMagnitudeOfVector(Ratio)
        else // just right
             Change.Clear;
        // Add Change to Accumulator
        if Flock <> nil then
           TFormationFlock(Flock).BoidApplyRule( Self, fbSeparation, Accumulator, Change );
        Accumulator.Add( Change );
     finally
        Change.Free;
     end;
end;

{ TFormationFlockCollection }

function TFormationFlockCollection.GetFlock: TFormationFlock;
begin
     result := TFormationFlock(inherited Flock);
end;

function TFormationFlockCollection.GetItem(Index: Integer): TFormationBoid;
begin
     result := TFormationBoid(inherited Items[Index]);
end;

procedure TFormationFlockCollection.SetItem(Index: Integer;
  const Value: TFormationBoid);
begin
     inherited Items[Index] := Value;
end;

{ TFormationFlock }

procedure TFormationFlock.AfterConstruction;
begin
     inherited AfterConstruction;
     Formation := ffLine;
end;

procedure TFormationFlock.BeginFormationUpdate;
begin
     Inc(FUpdateCount);
end;

constructor TFormationFlock.Create(AOwner: TComponent);
begin
     inherited Create(AOwner);
     // turn down strength separation in relation to
     // cohesion
     StrengthAlignment := 1;
     StrengthAvoidance := 1;
     StrengthSeparation := 0.2;
     SeparationDistance := 1;
     FDesiredPosition := TVector.Create;
     FDesiredVelocity := TVector.Create;
end;

function TFormationFlock.CreateFlock: TFlockCollection;
begin
     result := TFormationFlockCollection.Create( Self, TFormationBoid );
end;

procedure TFormationFlock.DefineFormation;
begin
     if (Boids = nil) or (Boids.Count = 0) then Exit;
     // set boids to formation
     case Formation of
          ffUserDefined:
            // user defined formation... if event not assigned then just normal
            // flocking behavior
            if Assigned(FOnDefineFormation) then FOnDefineFormation(Self);
          ffLine:       DefineLineFormation;
          ffSquare:     DefineSquareFormation;
          ffTwoSquares: DefineTwoSquaresFormation;
          ffWedge:      DefineWedgeFormation;
     end;
end;

procedure TFormationFlock.DefineLineFormation;
var
   i: Integer;
   Sep: Single;
   s: Single;
   v: TVector;
begin
     // Get the orthogonal vector to the group's orientation
     // Get the angle
     s := RadToDeg( ArcTan2( Velocity.Z, Velocity.X ) );
     // add 90 to angle
     s := s + 90;
     v := TVector.Create( Cos(DegToRad(s)), 0, Sin(DegToRad(s)) );
     try
        Sep := SeparationMagnitude*SeparationDistance * ( Boids.Count / 2 );
        // use separation distance
        for i := 0 to Boids.Count - 1 do
        begin
             Boids[i].Offset.X := V.X * Sep;
             Boids[i].Offset.Z := V.Z * sep;
             Sep := Sep - SeparationMagnitude*SeparationDistance;
        end;
     finally
        v.Free;
     end;
end;

procedure TFormationFlock.DefineSquare(Index, Count: Integer;
  const OffsetPos, Vel: TVector; Cols: Integer);
var
   Row, Col: Integer;
   i: Integer;
   SepCol: Single;
   SepRow: Single;
   v: TVector;
   s: Single;
begin
     // Get the orthogonal vector to the group's orientation
     // Get the angle
     s := RadToDeg( ArcTan2( Vel.Z, Vel.X ) );
     // add 90 to angle
     s := s + 90;
     v := TVector.Create( Cos(DegToRad(s)), 0, Sin(DegToRad(s)) );
     try
        V.Normalize;
        SepCol := SeparationMagnitude*SeparationDistance * ( Cols / 2 );
        Row := 0;
        SepRow := Row*SepCol;
        Col := 0;
        // use separation distance
        for i := Index to Count-1 do
        begin
             Boids[i].Offset.X := OffsetPos.X + v.X * SepCol - Vel.X*SepRow;
             Boids[i].Offset.Z := OffsetPos.Z + v.Z * SepCol - Vel.Z*SepRow;
             Inc(Col);
             if Col > Cols then
             begin
                  Col := 0;
                  Inc(Row);
                  SepCol := SeparationMagnitude*SeparationDistance*(Cols/2);
                  SepRow := Row*SepCol;
             end
             else
                 SepCol := SepCol - SeparationMagnitude*SeparationDistance;
        end;
     finally
        v.Free;
     end;
end;

procedure TFormationFlock.DefineSquareFormation;
var
   Cols: Integer;
   OffsetPos: TVector;
begin
     Cols := Round(Sqrt( Boids.Count ));
     OffsetPos := TVector.Create;
     try
        DefineSquare( 0, Boids.Count, OffsetPos, Velocity, Cols );
     finally
        OffsetPos.Free;
     end;
end;

procedure TFormationFlock.DefineTwoSquaresFormation;
var
   Cols: Integer;
   OffsetPos: TVector;
   v: TVector;
   s: Single;
begin
     // Get the orthogonal vector to the group's orientation
     // Get the angle
     s := RadToDeg( ArcTan2( Velocity.Z, Velocity.X ) );
     // add 90 to angle
     s := s + 90;
     v := TVector.Create( Cos(DegToRad(s)), 0, Sin(DegToRad(s)) );
     Cols := Round(Sqrt( Boids.Count div 2 ));
     OffsetPos := TVector.Create;
     try
        v.SetMagnitudeOfVector( 1 );
        OffsetPos.SetXYZ( v.x, 0, v.Z );
        OffsetPos.SetMagnitudeOfVector( 10 );
        DefineSquare( 0, Boids.Count div 2, OffsetPos, Velocity, Cols );
        OffsetPos.SetXYZ( v.x, 0, v.Z );
        OffsetPos.SetMagnitudeOfVector( -10 );
        DefineSquare( (Boids.Count div 2), Boids.Count, OffsetPos, Velocity, Cols );
     finally
        OffsetPos.Free;
        v.Free;
     end;
end;

procedure TFormationFlock.DefineWedgeFormation;
var
   i: Integer;
   Sep: Single;
   s1, s2: Single;
   v1, v2: TVector;
begin
     // Get the 120 Degrees vector to the group's orientation
     // Get the angle
     s1 := RadToDeg( ArcTan2( Velocity.Z, Velocity.X ) );
     // add 120 to angle
     s2 := s1 - 120;
     s1 := s1 + 120;
     v1 := TVector.Create( Cos(DegToRad(s1)), 0, Sin(DegToRad(s1)) );
     v2 := TVector.Create( Cos(DegToRad(s2)), 0, Sin(DegToRad(s2)) );
     try
        Sep := SeparationMagnitude*SeparationDistance * ( Boids.Count / 2 );
        // use separation distance
        for i := 0 to (Boids.Count div 2) do
        begin
             Boids[i].Offset.X := V1.X * Sep;
             Boids[i].Offset.Z := V1.Z * sep;
             Boids[Boids.Count - 1 - i].OffSet.X := v2.X * Sep;
             Boids[Boids.Count - 1 - i].OffSet.Z := v2.Z * Sep;
             Sep := Sep - SeparationMagnitude*SeparationDistance;
        end;
        if (Boids.Count > 1) and
           ((Boids.Count mod 2) <> 0) then
        begin
             Boids[Boids.Count div 2 + 1].Offset.X := 0;
             Boids[Boids.Count div 2 + 1].Offset.Z := 0;
        end;
     finally
        v1.Free;
        v2.Free;
     end;
end;

procedure TFormationFlock.EndFormationUpdate;
begin
     Dec(FUpdateCount);
     FormationChange;
end;

procedure TFormationFlock.FormationChange;
begin
     if FUpdateCount = 0 then DefineFormation;
end;

function TFormationFlock.GetBoids: TFormationFlockCollection;
begin
     result := TFormationFlockCollection(inherited Boids);
end;

procedure TFormationFlock.SetBoids(const Value: TFormationFlockCollection);
begin
     inherited Boids := Value;
end;

procedure TFormationFlock.SetFormation(const Value: TFlockFormation);
begin
     if Formation <> Value then
     begin
          FFormation := Value;
          DefineFormation;
     end;
end;

procedure TFormationFlock.Execute;
var
   i: Integer;
   Change: TVector;
   Dist: Single;
begin
  BeginUpdate;
  try
     Executing;
     // if desired velocity is different than velocity then move towards it
     if not DesiredVelocity.EqualsStar( Velocity ) then
     begin
          Change := TVector.Create( DesiredVelocity );
          try
             Change.Subtract( Velocity );
             // Really Need to change this to interpolate so that the velocity
             // doesn't keep oscillating
             Change.Multiply(MaxChange/3);
             Velocity.Add(Change);
          finally
             Change.Free;
          end;
     end;
     // if desired position is different AND velocity = DesiredVelocity then MOVE
     if (not DesiredPosition.EqualsStar(Position)) and
        DesiredVelocity.EqualsStar( Velocity ) then
     begin
          Change := TVector.Create( DesiredPosition );
          try
             Change.Subtract(Position);
             // if we are almost to our destination, start slowing down
             Dist := DesiredPosition.DistanceTo( Position );
             if (Dist/MaxChange) < MinTimeToMaxSpeed then
                 Change.Multiply(MaxSpeed*0.80-(MaxChange/Dist))
             else
                 // go slower than maxspeed to allow each boid to dress up their
                 // position in the formation
                 Change.SetMagnitudeOfVector(MaxSpeed*0.80);
             Position.Add(Change);
          finally
             Change.Free;
          end;
     end;
     // Update every member of the flock for the time period
     for i := 0 to Boids.Count - 1 do
         Boids[i].Execute;
     Executed;
  finally
     EndUpdate;
  end;
end;

procedure TFormationFlock.Assign(Source: TPersistent);
begin
     if Source is TFormationFlock then
     begin
          BeginFormationUpdate;
          try
             FFormation := TFormationFlock(Source).Formation;
             inherited Assign(Source);
          finally
             EndFormationUpdate;
          end;
     end
     else
         inherited Assign(Source);
end;

destructor TFormationFlock.Destroy;
begin
     FDesiredPosition.Free;
     FDesiredVelocity.Free;
     // prevent reformation
     BeginFormationUpdate;
     inherited Destroy;
end;

procedure TFormationFlock.SetDesiredPosition(const Value: TVector);
begin
     FDesiredPosition.Assign(Value);
end;

procedure TFormationFlock.SetDesiredVelocity(const Value: TVector);
begin
     FDesiredVelocity.Assign(Value);
end;

procedure TFormationFlock.VectorChange( Sender: TVector );
begin
     inherited VectorChange( Sender );
     if Sender = Velocity then
        FormationChange;
end;

procedure TFormationFlock.SetSeparationDistance(const Value: Single);
begin
     if Value <> SeparationDistance then
     begin
          inherited SetSeparationDistance(Value);
          FormationChange;
     end;
end;

end.
